@@ -1,4 +1,4 @@
-# Copyright 2024 The Orbax Authors.
+# Copyright 2025 The Orbax Authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -19,12 +19,13 @@ not leave users free to define their own temporary path structure. The current
 implementation is mainly a refactoring of old logic that separately created
 temp directories and finalized them. It does not touch other logic that detects
 temp checkpoints and cleans them up (primarily located in
-orbax.checkpoint.path.step and CheckpointManager).
+orbax.checkpoint.path.step and :py:class:`.CheckpointManager`).
 
-Ordinarily, atomic logic defaults to `AtomicRenameTemporaryPath`, which uses an
-atomic rename to indicate checkpoint completion. However, not all filesystems
-support atomic rename, so `CommitFileTemporaryPath` is provided as an
-alternative, which uses a "commit_success" file to indicate completion.
+Ordinarily, atomic logic defaults to :py:class:`AtomicRenameTemporaryPath`,
+which uses an atomic rename to indicate checkpoint completion. However, not all
+filesystems support atomic rename, so :py:class:`CommitFileTemporaryPath` is
+provided as an alternative, which uses a "commit_success" file to indicate
+completion.
 
 Ideally, we would standardize on a single behavior, but it is difficult, largely
 for legacy reasons, to achieve this. Furthermore, there are many other
@@ -67,7 +68,6 @@ from orbax.checkpoint._src.futures import synchronization
 from orbax.checkpoint._src.metadata import checkpoint as checkpoint_metadata
 from orbax.checkpoint._src.metadata import step_metadata_serialization
 from orbax.checkpoint._src.multihost import multihost
-from orbax.checkpoint._src.path import async_utils
 from orbax.checkpoint._src.path import atomicity_types
 from orbax.checkpoint._src.path import step as step_lib
 from orbax.checkpoint._src.path import utils
@@ -76,10 +76,25 @@ from orbax.checkpoint._src.path import utils
 TMP_DIR_SUFFIX = step_lib.TMP_DIR_SUFFIX
 COMMIT_SUCCESS_FILE = step_lib._COMMIT_SUCCESS_FILE  # pylint: disable=protected-access
 
-
 _module_unique_count = itertools.count()
 
 
+async def _mkdir(path: epath.Path, *args, **kwargs):
+  return await asyncio.to_thread(path.mkdir, *args, **kwargs)
+
+
+async def _exists(path: epath.Path):
+  return await asyncio.to_thread(path.exists)
+
+
+async def _is_tmp_checkpoint(path: epath.Path):
+  return await asyncio.to_thread(step_lib.is_tmp_checkpoint, path)
+
+
+async def _rmtree(path: epath.Path):
+  return await asyncio.to_thread(path.rmtree)
+
+
 class AsyncMakeDirFunc(Protocol):
 
   def __call__(
@@ -128,14 +143,14 @@ async def _create_tmp_directory(
     FileExistsError: if tmp directory already exists.
   """
   if multihost.is_primary_host(primary_host):
-    if await async_utils.async_exists(tmp_dir):
-      if await async_utils.async_is_tmp_checkpoint(tmp_dir):
+    if await _exists(tmp_dir):
+      if await _is_tmp_checkpoint(tmp_dir):
         logging.warning(
             'Attempted to create temporary directory %s which already exists.'
             ' Removing existing directory since it is not finalized.',
             tmp_dir,
         )
-        await async_utils.async_rmtree(tmp_dir)
+        await _rmtree(tmp_dir)
       else:
         raise FileExistsError(
             f'Attempted to create temporary directory {tmp_dir} which already'
@@ -271,7 +286,7 @@ class AtomicRenameTemporaryPath(atomicity_types.TemporaryPath):
         file_options.path_permission_mode or self._path_permission_mode or None
     )
     return await _create_tmp_directory(
-        async_utils.async_makedirs,
+        _mkdir,
         self._tmp_path,
         primary_host=self._primary_host,
         path_permission_mode=mode,
@@ -281,7 +296,7 @@ class AtomicRenameTemporaryPath(atomicity_types.TemporaryPath):
   def finalize(
       self,
   ):
-    """Finalizes atomic save by renaming tmp_dir or writing a success file.
+    """Finalizes atomic save by renaming tmp_dir.
 
     Updates checkpoint metadata with commit_timestamp_nsecs.
 
@@ -395,7 +410,7 @@ class CommitFileTemporaryPath(atomicity_types.TemporaryPath):
         file_options.path_permission_mode or self._path_permission_mode or None
     )
     return await _create_tmp_directory(
-        async_utils.async_makedirs,
+        _mkdir,
         self._tmp_path,
         primary_host=self._primary_host,
         path_permission_mode=mode,
@@ -405,7 +420,7 @@ class CommitFileTemporaryPath(atomicity_types.TemporaryPath):
   def finalize(
       self,
   ):
-    """Finalizes atomic save by renaming tmp_dir or writing a success file.
+    """Finalizes atomic save by writing a success file.
 
     Updates checkpoint metadata with commit_timestamp_nsecs.
 
@@ -475,6 +490,7 @@ def create_all_async(
     multiprocessing_options: Optional[
         options_lib.MultiprocessingOptions
     ] = None,
+    subdirectories: Sequence[str] | None = None,
 ) -> future.Future:
   """Creates all temporary paths in parallel asynchronously.
 
@@ -484,6 +500,9 @@ def create_all_async(
       Also adds them to the awaitable signals contract.
     multiprocessing_options: MultiprocessingOptions to use for barrier syncs and
       primary host.
+    subdirectories: Sequence of subdirectories to create under `paths`. If not
+      provided, no subdirectories will be created. The same set of
+      subdirectories will be created under each path in `paths`.
 
   Returns:
     A future that which sends the completion signals when all paths are created.
@@ -508,7 +527,11 @@ def create_all_async(
   commit_future = future.NoopFuture()
   if multihost.is_primary_host(primary_host):
     commit_future = future.CommitFutureAwaitingContractedSignals(
-        _create_paths(paths),
+        _create_paths(
+            paths,
+            subdirectories=subdirectories,
+            multiprocessing_options=multiprocessing_options,
+        ),
         send_signals=completion_signals,
         timeout_secs=multihost.DIRECTORY_CREATION_TIMEOUT,
     )
@@ -527,11 +550,24 @@ def create_all_async(
 
 
 async def _create_paths(
-    paths: Sequence[atomicity_types.TemporaryPath],
+    tmp_paths: Sequence[atomicity_types.TemporaryPath],
+    subdirectories: Sequence[str] | None = None,
+    *,
+    multiprocessing_options: options_lib.MultiprocessingOptions,
 ):
   """Creates all temporary paths in parallel."""
   start = time.time()
-  await asyncio.gather(*[path.create() for path in paths])
+  paths = await asyncio.gather(*[path.create() for path in tmp_paths])
+  if subdirectories and multihost.is_primary_host(
+      multiprocessing_options.primary_host
+  ):
+    creation_ops = []
+    for path in paths:
+      creation_ops.extend([
+          _mkdir(path / name, parents=False, exist_ok=False)
+          for name in subdirectories
+      ])
+    await asyncio.gather(*creation_ops)
   directory_creation_secs = time.time() - start
   jax.monitoring.record_event_duration_secs(
       '/jax/orbax/write/directory_creation_secs',
