@@ -1,4 +1,4 @@
-# Copyright 2024 The Orbax Authors.
+# Copyright 2025 The Orbax Authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -36,10 +36,12 @@ from orbax.checkpoint import options as options_lib
 from orbax.checkpoint import utils
 from orbax.checkpoint._src import threading as threading_lib
 from orbax.checkpoint._src.checkpoint_managers import policy_checkpoint_info
+from orbax.checkpoint._src.checkpoint_managers import preservation_policy as preservation_policy_lib
 from orbax.checkpoint._src.checkpoint_managers import save_decision_policy as save_decision_policy_lib
 from orbax.checkpoint._src.checkpointers import abstract_checkpointer
 from orbax.checkpoint._src.checkpointers import async_checkpointer
 from orbax.checkpoint._src.checkpointers import checkpointer as checkpointer_lib
+from orbax.checkpoint._src.futures import synchronization
 from orbax.checkpoint._src.handlers import checkpoint_handler
 from orbax.checkpoint._src.handlers import composite_checkpoint_handler
 from orbax.checkpoint._src.handlers import handler_registration
@@ -206,6 +208,36 @@ def _get_default_save_decision_policy(
   return save_decision_policy_lib.AnySavePolicy(save_interval_policies)
 
 
+def _get_default_preservation_policy(
+    options: CheckpointManagerOptions,
+) -> preservation_policy_lib.PreservationPolicy:
+  """Returns a default preservation policy."""
+  preservation_policies = []
+  if options.keep_period is not None:
+    preservation_policies.append(
+        preservation_policy_lib.EveryNSteps(options.keep_period)
+    )
+  if options.keep_time_interval is not None:
+    total_seconds = int(options.keep_time_interval.total_seconds())
+    preservation_policies.append(
+        preservation_policy_lib.EveryNSeconds(interval_secs=total_seconds)
+    )
+  if options.best_fn is not None:
+    preservation_policies.append(
+        preservation_policy_lib.BestN(
+            get_metric_fn=options.best_fn,
+            reverse=(options.best_mode == 'min'),
+            n=options.max_to_keep,
+            keep_checkpoints_without_metrics=options.keep_checkpoints_without_metrics,
+        )
+    )
+  else:
+    preservation_policies.append(
+        preservation_policy_lib.LatestN(n=options.max_to_keep)
+    )
+  return preservation_policy_lib.AnyPreservationPolicy(preservation_policies)
+
+
 # TODO(b/268051457) Clean up when no longer depended upon by internal users.
 def is_async_checkpointer(checkpointer: AbstractCheckpointer):
   return isinstance(
@@ -219,35 +251,19 @@ def is_async_checkpointer(checkpointer: AbstractCheckpointer):
 # TODO(b/309965339) Set todelete_subdir defaults if directory is on CNS.
 @dataclasses.dataclass
 class CheckpointManagerOptions:
-  """Optional arguments for CheckpointManager.
+  """Optional arguments for :py:class:`.CheckpointManager`.
 
   save_interval_steps:
     The interval at which checkpoints should be saved.
     Ensures checkpoints will only be saved every n steps. Defaults to 1.
   max_to_keep:
-    If provided, specifies the maximum number of checkpoints to
-    keep. Older checkpoints are removed. By default, does not remove any old
-    checkpoints. Must be None or non-negative. When set, checkpoints
-    may be considered for deletion when there are more than `max_to_keep`
-    checkpoints present. Checkpoints are kept if they meet any of the conditions
-    below, such as `keep_time_interval`, `keep_period`, etc. Any remaining
-    checkpoints that do not meet these conditions are garbage-collected.
+    deprecated, do not use. use `preservation_policy` instead.
   keep_time_interval:
-    When more than max_to_keep checkpoints are present,
-    an older checkpoint that would ordinarily be deleted will be preserved if it
-    has been at least `keep_time_interval` since the previous preserved
-    checkpoint. The default setting of `None` does not preserve any checkpoints
-    in this way. For example, this may be used to ensure checkpoints are
-    retained at a frequency of approximately than one per hour.
+    deprecated, do not use. use `preservation_policy` instead.
   keep_period:
-    If set, any existing checkpoints matching checkpoint_step % keep_period == 0
-    will not be deleted.
+    deprecated, do not use. use `preservation_policy` instead.
   should_keep_fn:
-    WARNING: It is an experimental feature and may change without notice.
-    If set then this Callable overrides the behavior of `keep_period`, which is
-    ignored. It does not change the behavior of `keep_time_interval` and
-    `keep_checkpoints_without_metrics`. It is a predicate with signature:
-    `Callable[[int], bool]`, where int is the step number of the checkpoint.
+    deprecated, do not use. use `preservation_policy` instead.
   best_fn:
     If set, maintains checkpoints based on the quality of given
     metrics rather than recency. The function should accept a PyTree of metrics,
@@ -311,14 +327,20 @@ class CheckpointManagerOptions:
   save_root_metadata: If True, saves root-level metadata about checkpoints.
     This metadata is not step-specific and is written only once.
   temporary_path_class:
-    Optional. The concrete `atomicity_types.TemporaryPath` class to be used by
-    the underlying `Checkpointer`.
+    Optional. The concrete :py:class:`.atomicity_types.TemporaryPath` class to
+    be used by the underlying :py:class:`.Checkpointer`.
   save_decision_policy: An object used to determine when a checkpoint should be
     saved. If provided, overrides any other options dealing with this subject,
     including `save_interval_steps`, `save_on_steps`, and `should_save_fn`, and
     is the sole means of determining when a checkpoint should be saved. If not
     provided, these other options are used instead. Prefer to use this option
     over others.
+  preservation_policy: An object used to determine which checkpoints to
+    preserve. If provided, overrides any other options dealing with this
+    subject, including `max_to_keep`, `keep_time_interval`, `keep_period`, and
+    `should_keep_fn`, `best_fn`, and is the sole means of determining which
+    checkpoints to preserve. If not provided, these other options are used
+    instead. Prefer to use this option over others.
   prevent_write_metrics: False by default. If True, metrics will not be written.
   """
 
@@ -352,6 +374,9 @@ class CheckpointManagerOptions:
   save_decision_policy: Optional[
       save_decision_policy_lib.SaveDecisionPolicy
   ] = None
+  preservation_policy: Optional[preservation_policy_lib.PreservationPolicy] = (
+      None
+  )
   prevent_write_metrics: bool = False
 
   def __post_init__(self):
@@ -382,7 +407,11 @@ class CheckpointManagerOptions:
           f"or 'max'. Got {self.best_mode}."
       )
       raise ValueError(msg)
-    if self.max_to_keep is not None and self.max_to_keep < 0:
+    if (
+        self.preservation_policy is None
+        and self.max_to_keep is not None
+        and self.max_to_keep < 0
+    ):
       raise ValueError('Setting of `max_to_keep` must be None or non-negative.')
     if self.read_only and self.save_interval_steps > 0:
       self.save_interval_steps = 0
@@ -436,7 +465,7 @@ class CheckpointManagerOptions:
           'CheckpointManagerOptions.read_only=True, setting'
           ' should_save_fn=None.'
       )
-    if self.should_keep_fn is not None:
+    if self.preservation_policy is None and self.should_keep_fn is not None:
       logging.warning(
           'CheckpointManagerOptions.should_keep_fn is set, setting'
           ' keep_period=None (was %s).',
@@ -532,10 +561,11 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
     https://orbax.readthedocs.io/en/latest/guides/checkpoint/api_refactor.html
     for technical details.
 
-    The `CheckpointManager` is ultimately backed by a single `Checkpointer`, to
-    which saving and restoring is delegated. Behind step management options,
-    metrics-related logic, and other frills, saving and restoring with
-    `CheckpointManager` is quite similar to using
+    The `CheckpointManager` is ultimately backed by a single
+    :py:class:`.Checkpointer`, to which saving and restoring is delegated.
+    Behind
+    step management options, metrics-related logic, and other frills, saving and
+    restoring with `CheckpointManager` is quite similar to using
     `Checkpointer(CompositeCheckpointHandler)`.
 
     Example::
@@ -635,6 +665,22 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
         or _get_default_save_decision_policy(self._options)
     )
 
+    if self._options.preservation_policy and (
+        self._options.keep_time_interval
+        or self._options.max_to_keep
+        or self._options.keep_period
+        or self._options.should_keep_fn
+    ):
+      raise ValueError(
+          '`preservation_policy` and `delete options` are mutually exclusive'
+          ' - do not use together.'
+      )
+
+    self._preservation_policy = (
+        self._options.preservation_policy
+        or _get_default_preservation_policy(self._options)
+    )
+
     if self._options.best_mode not in ['min', 'max']:
       raise ValueError('`best_mode` must be one of: "min", "max"')
 
@@ -743,6 +789,9 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
       )
 
     self._directory = epath.Path(directory)
+    self._save_tracker = synchronization.OpTrackerFactory.create_tracker(
+        'checkpoint_manager_save'
+    )
     if self._options.read_only:
       logging.warning('Given directory is read only=%s', self._directory)
     if self._options.create:
@@ -1321,7 +1370,10 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
         '/jax/checkpoint/write/wait_for_prev_duration_secs',
         step_stats.wait_for_prev_duration_secs,
     )
-
+    self._save_tracker = synchronization.OpTrackerFactory.create_tracker(
+        'checkpoint_manager_save'
+    )
+    self._save_tracker.start()
     if step in self.all_steps():
       raise StepAlreadyExistsError(
           f'Checkpoint for step {step} already exists.'
@@ -1445,7 +1497,10 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
     if is_async_checkpointer(self._checkpointer):
 
       def launch_finalize_thread() -> _FinalizeThread:
-        assert not self.is_saving_in_progress(), (
+        is_finalize_in_progress = self._finalize_thread.map(
+            lambda t: t is not None and t.is_alive()
+        )
+        assert not is_finalize_in_progress, (
             'Save finalization already in progress for'
             f' step={self._finalize_thread.get_not_none().step()}'
         )
@@ -1641,19 +1696,7 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
 
   # TODO(b/370812224): Deprecate in favor of StepMetadata.metrics
   def metrics(self, step: int) -> Optional[PyTree]:
-    try:
-      # Use handler directly, since this happens in a background thread and
-      # barriers cannot be used. This usage pattern is not
-      # recommended in other contexts.
-      metrics = self._metrics_handler.restore(
-          self._get_read_step_directory(step, self.directory)
-          / METRIC_ITEM_NAME
-      )
-      return metrics
-    except FileNotFoundError as e:
-      logging.warning('Missing metrics for step %d', step)
-      logging.error(e)
-      return None
+    return self._get_metrics(step)
 
   @property
   def _metrics_handler(self) -> CheckpointHandler:
@@ -1676,9 +1719,7 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
       a list of CheckpointInfo, sorted by increasing step.
     """
     if not self.directory.exists():
-      raise ValueError(
-          f'Checkpoint root directory {self.directory} does not exist.'
-      )
+      return []
     start = time.time()
     step_metadatas = self._step_name_format.find_all(self.directory)
 
@@ -1705,22 +1746,6 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
       )
       return checkpoint_infos
 
-  def _get_interval_preserved_checkpoints(
-      self, checkpoints: checkpoint_info.CheckpointInfos
-  ) -> List[CheckpointInfo]:
-    """Gets which checkpoints should be kept based on keep_time_interval."""
-    if checkpoints.empty():
-      return []
-    interval_preserved_checkpoints = [checkpoints[0]]
-    if self._options.keep_time_interval is not None:
-      for info in checkpoints[1:]:
-        if info.time >= (
-            interval_preserved_checkpoints[-1].time
-            + self._options.keep_time_interval
-        ):
-          interval_preserved_checkpoints.append(info)
-    return interval_preserved_checkpoints
-
   def _add_checkpoint_info(self, step: int, metrics: Optional[PyTree]):
     self._checkpoints.append(
         CheckpointInfo(
@@ -1763,18 +1788,25 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
           processes=self._multiprocessing_options.active_processes,
       )
 
-  def _get_step_metadata(self, step: int) -> StepMetadata:
-    infos = [info for info in self._checkpoints if info.step == step]
-    if not infos:
-      metrics = None
-    else:
-      if len(infos) > 1:
-        logging.warning(
-            'Multiple CheckpointInfos found for step %d. Using the first one.',
-            step,
-        )
-      metrics = infos[0].metrics
+  def _get_metrics(self, step: int) -> PyTree | None:
+    try:
+      # Use handler directly, since this happens in a background thread and
+      # barriers cannot be used. This usage pattern is not
+      # recommended in other contexts.
+      metrics = self._metrics_handler.restore(
+          self._get_read_step_directory(step, self.directory) / METRIC_ITEM_NAME
+      )
+      return metrics
+    except FileNotFoundError:
+      logging.log_first_n(
+          logging.INFO,
+          'Missing metrics for step %d',
+          1,  # log only once
+          step,
+      )
+      return None
 
+  def _get_step_metadata(self, step: int) -> StepMetadata:
     step_metadata = self._checkpointer.metadata(
         self._get_read_step_directory(step, self.directory),
     )
@@ -1788,6 +1820,7 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
         step_metadata.item_metadata
     )
 
+    metrics = self._get_metrics(step)
     if metrics is not None:
       validated_metrics = step_metadata_serialization.deserialize(
           {}, metrics=dict(metrics)
@@ -1867,102 +1900,27 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
 
   def _get_old_steps_to_remove(self) -> List[int]:
     """Returns checkpoints that should be deleted."""
-    # Must have set max_to_keep in order to remove any checkpoints.
-    if self._options.max_to_keep is None:
-      return []
-    # Not enough checkpoints accumulated to consider deletion.
-    if self._checkpoints.size() <= self._options.max_to_keep:
-      return []
-
     # This isn't a duration but there isn't a general counter that we can use so
     # we abuse a duration metric to count the number of steps examined.
     jax.monitoring.record_event_duration_secs(
         '/jax/checkpoint/write/old_steps_examined_count',
         self._checkpoints.size(),
     )
-
-    if self._track_best:
-      # Best steps (to keep) are at the end, after sorting.
-      (
-          checkpoints_without_metrics,
-          sorted_checkpoints,
-      ) = self._sort_checkpoints_by_metrics(self._checkpoints)
-    else:
-      # checkpoints already sorted by ascending step
-      checkpoints_without_metrics = []
-      sorted_checkpoints = [info for info in self._checkpoints]
-
-    keep = int(self._options.max_to_keep)
-    if self._options.keep_checkpoints_without_metrics:
-      maybe_delete = (
-          sorted_checkpoints[:-keep] if keep > 0 else sorted_checkpoints
-      )
-      active_checkpoints = set(
-          checkpoints_without_metrics
-          + (sorted_checkpoints[-keep:] if keep > 0 else [])
-      )
-    else:
-      all_checkpoints = checkpoints_without_metrics + sorted_checkpoints
-      maybe_delete = all_checkpoints[:-keep] if keep > 0 else sorted_checkpoints
-      active_checkpoints = set(all_checkpoints[-keep:] if keep > 0 else [])
-
-    interval_preserved_checkpoints = self._get_interval_preserved_checkpoints(
-        self._checkpoints
+    preservation_result = self._preservation_policy.should_preserve(
+        [info for info in self._checkpoints],
+        context=preservation_policy_lib.PreservationContext(),
     )
-    kept_checkpoints = set()
-    for info in maybe_delete:
-      if (
-          self._options.keep_time_interval is not None
-          and interval_preserved_checkpoints
-      ):
-        if info in interval_preserved_checkpoints:
-          logging.info(
-              'Preserving %s: (Reason: older falling on keep_time_interval).',
-              info,
-          )
-          kept_checkpoints.add(info)
-          continue
-        elif info.time >= (
-            interval_preserved_checkpoints[-1].time
-            + self._options.keep_time_interval
-        ):
-          interval_preserved_checkpoints.append(info)
-          logging.info(
-              'Preserving %s: (Reason: latest falling on keep_time_interval).',
-              info,
-          )
-          kept_checkpoints.add(info)
-          continue
-      if (
-          self._options.should_keep_fn is not None
-          and self._options.should_keep_fn(info.step)
-      ):
-        logging.info(
-            'Preserving %s: (Reason: on should_keep_fn callback).', info
+    return [
+        checkpoint.step
+        for checkpoint, should_preserve in zip(
+            self._checkpoints, preservation_result
         )
-        kept_checkpoints.add(info)
-        continue
-      if (
-          self._options.keep_period is not None
-          and info.step % self._options.keep_period == 0
-      ):
-        logging.info(
-            'Preserving %s: (Reason: on keep_period=%s).',
-            info,
-            self._options.keep_period,
+        if not should_preserve
+        and (
+            self._options.should_keep_fn is None
+            or not self._options.should_keep_fn(checkpoint.step)
         )
-        kept_checkpoints.add(info)
-        continue
-
-    kept_checkpoints.update(active_checkpoints)
-
-    steps_to_remove = []
-    for info in self._checkpoints:
-      if info not in kept_checkpoints:
-        reason = 'worse metric' if self._track_best else 'old checkpoint'
-        logging.info('Deleting %s: (Reason: %s).', info, reason)
-        steps_to_remove.append(info.step)
-    return steps_to_remove
+    ]
 
   def _wait_for_checkpointers(self):
     if is_async_checkpointer(self._checkpointer):
@@ -1975,8 +1933,8 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
     synchronous, since old checkpoints are already cleaned up immediately after
     completing `save`, and there is no background thread to wait for.
 
-    If some checkpointers are of type AsyncCheckpointer, however, this method
-    will wait until each of these checkpointers is finished.
+    If some checkpointers are of type :py:class:`.AsyncCheckpointer`, however,
+    this method will wait until each of these checkpointers is finished.
     """
     process_index = multihost.process_index()
     current_thread = threading.current_thread()
@@ -2030,7 +1988,14 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
 
   def is_saving_in_progress(self) -> bool:
     """Returns whether a checkpoint save is in progress."""
-    return self._finalize_thread.map(lambda t: t is not None and t.is_alive())
+    processes_saving = self._save_tracker.get_in_progress_ids()
+    logging.vlog(
+        1,
+        '[process=%s][is_saving_in_progress] Processes saving: %s',
+        multihost.process_index(),
+        processes_saving,
+    )
+    return bool(processes_saving)
 
   def check_for_errors(self):
     """Checks for any outstanding errors in completed asynchronous save operations.
@@ -2081,39 +2046,42 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
     """Finalizes individual items and starts garbage collection."""
     process_index = multihost.process_index()
     current_thread = threading.current_thread()
-    self._non_blocking_metadata_store.wait_until_finished()
-    self._wait_for_checkpointers()
-    # If an error is encountered while waiting for commit futures to complete,
-    # we will not proceed past this point.
-    self._finalize_checkpoint(step)
-    remove_steps_start_time = time.time()
-    self._checkpoint_deleter.delete_steps(steps_to_remove)
-    jax.monitoring.record_event_duration_secs(
-        '/jax/checkpoint/write/remove_steps_duration_secs',
-        time.time() - remove_steps_start_time,
-    )
-    logging.info(
-        '[process=%s][thread=%s][step=%s] CheckpointManager Save Finalize is'
-        ' syncing with other hosts...',
-        process_index,
-        current_thread.name,
-        step,
-    )
-    barrier_sync_fn = self._create_thread_safe_barrier_sync_fn()
-    barrier_sync_fn(
-        multihost.unique_barrier_key(
-            'CheckpointManager:finalize',
-            prefix=self._multiprocessing_options.barrier_sync_key_prefix,
-            suffix=str(step),
-        )
-    )
-    logging.info(
-        '[process=%s][thread=%s][step=%s] CheckpointManager Save Finalize is'
-        ' done on all hosts.',
-        process_index,
-        current_thread.name,
-        step,
-    )
+    try:
+      self._non_blocking_metadata_store.wait_until_finished()
+      self._wait_for_checkpointers()
+      # If an error is encountered while waiting for commit futures to complete,
+      # we will not proceed past this point.
+      self._finalize_checkpoint(step)
+      remove_steps_start_time = time.time()
+      self._checkpoint_deleter.delete_steps(steps_to_remove)
+      jax.monitoring.record_event_duration_secs(
+          '/jax/checkpoint/write/remove_steps_duration_secs',
+          time.time() - remove_steps_start_time,
+      )
+    finally:
+      self._save_tracker.complete()
+      logging.info(
+          '[process=%s][thread=%s][step=%s] CheckpointManager Save Finalize is'
+          ' syncing with other hosts...',
+          process_index,
+          current_thread.name,
+          step,
+      )
+      barrier_sync_fn = self._create_thread_safe_barrier_sync_fn()
+      barrier_sync_fn(
+          multihost.unique_barrier_key(
+              'CheckpointManager:finalize',
+              prefix=self._multiprocessing_options.barrier_sync_key_prefix,
+              suffix=str(step),
+          )
+      )
+      logging.info(
+          '[process=%s][thread=%s][step=%s] CheckpointManager Save Finalize is'
+          ' done on all hosts.',
+          process_index,
+          current_thread.name,
+          step,
+      )
 
   def close(self):
     """Waits for outstanding operations to finish and closes internal objects."""
